//==================================================================================================
// Name        : caLab.cpp
// Author      : Tobias Hft, Carsten Winkler
// Version     : 1.0.0.2
// Copyright   : HZB
// Description : library for get, put and monitoring EPICS variables (process variables) in LabVIEW
//==================================================================================================
#include "caLab.h"

#include <windows.h>
#include <iostream>
#include <stdio.h>
#include <stdlib.h>
#include <cmath>

#include <alarmString.h>
#include <epicsTime.h>
#include <db_access.h>
#include <epicsThread.h>

using namespace std;

void preload() {
	return;
}

//----------------------------------------------------------------------------
// Function:	initalizeVariable
//
// Description:		Create a channel for a process variable and allocate
// 					a structure ('Admin') to manage the data.
//
// Arg(s) In:	pvname - name of the process variable to initialize
//
// Arg(s) Out:	sError - error message
//				id - channel id according to pvname
//
// Return:		Error code
//----------------------------------------------------------------------------
int initalizeVariable(	char sError[MAX_ERROR_SIZE],
						char pvname[MAX_STRING_SIZE],
						long* id) {
	epicsMutexLock(mutex);
	// check CA Context
	if(!isRunning) {
		wsprintf(sError, "Channel access context not yet ready.");
		epicsMutexUnlock(mutex);
		return 6615;
	}
	// define variables
	int iResult = -1;
	unsigned int i; // counter
	// check valid process name
	if(*pvname == 0 || strlen(pvname) > MAX_STRING_SIZE || strlen(pvname) == 0) {
		wsprintf(sError, "Name of process variable is invalid. The max. length is %i.", MAX_STRING_SIZE);
		epicsMutexUnlock(mutex);
		return 6600;
	}
	// ignore already added PVs
	for(i = 0; i < admin.size(); i++) {
		if(strcmp(admin.at(i)->name, pvname) == 0) {
			// copy id
			*id = (long)admin.at(i)->chanid;
			// write error
			wsprintf(sError, "%s already initalized", pvname);
			epicsMutexUnlock(mutex);
			return 6601;
		}
	}
	// create ca channel and store informations for administration
	chid chanid = 0;
	// add "connection changed" listener for PV
	iResult = ca_create_channel(pvname, connectionChanged, NULL, CA_PRIORITY_DEFAULT, &chanid);
	// ca_pend_io will not block waiting for the channel to enter a connected state, because connectionHandler is defined
	i = 0;
	while((ca_state(chanid) != cs_conn) && (i < 9)) {
		i++;
		ca_pend_event(0.1);
	}
	if(ca_state(chanid) != cs_conn) {
		// write error message
		wsprintf(sError, "Timeout error for %s", pvname);
		epicsMutexUnlock(mutex);
		return 6619;
	}
	// write channel id
	*id = (long)chanid;
	// check execution of ca_create_channel()
	if(iResult != ECA_NORMAL) {
		// write error message
		wsprintf(sError, "%s occurred while creating channel for %s", ca_message(iResult), pvname);
		epicsMutexUnlock(mutex);
		return 6602;
	}
	// store handles for administration
	Admin* ad = (Admin*)malloc(sizeof(Admin));
	// check allocation
	if(!ad) {
		wsprintf(sError, "Allocation error for %s", pvname);
		epicsMutexUnlock(mutex);
		return 6612;
	}
	strcpy(ad->name, pvname);
	// set default values
	ad->connectionStatus = cs_closed;
	ad->elems = 0;
	ad->chanid = chanid;
	ad->eventid = 0;
	ad->eventidEnum = 0;
	ad->dbrType = DBF_NO_ACCESS;
	ad->monitor = 0;
	ad->enumRecord = 0;
	ad->timestamp.nsec = 0;
	ad->timestamp.secPastEpoch = 0;
	ad->sValueArr = NULL;
	ad->dValueArr = NULL;
	*(ad->sStatus) = 0;
	ad->dStatus = 0;
	*(ad->sSeverity) = 0;
	ad->dSeverity = 0;
	*(ad->sTime) = 0;
	ad->dTime = 0;
	*(ad->sError) = 0;
	ad->dError = 0;
	ad->putArr = NULL;
	ad->elemsToPut = 0;
	ad->chanidEgu = 0;
	ad->chanidDesc = 0;
	*(ad->unit) = 0;
	*(ad->desc) = 0;
	// store to admin list
	admin.push_back(ad);

	// created channel successfully
	epicsMutexUnlock(mutex);
	return 0;
}

//----------------------------------------------------------------------------
// Function:	connectionChanged
//
// Description:		Callback function for connection events.
// 					Checks the status of the connection.
//					Starts a monitor and write values (caput) if requested.
//
// Arg(s) In:	connection_handler_args (see CA manual)
//----------------------------------------------------------------------------
void connectionChanged(struct connection_handler_args args) {
	epicsMutexLock(mutex);
	// check CA Context
	if(!isRunning) {
		epicsMutexUnlock(mutex);
		return;
	}
	// define variables
	int iResult = -1;
	unsigned int i; // counter
	chtype dbrType;
	chtype dbrTypeEnum = DBR_TIME_SHORT;
	unsigned long elems;
	chid chanid = args.chid;
	// search PV in vector list
	for(i = 0; i < admin.size(); i++) {
		if(admin.at(i)->chanid == chanid) {
			// found PV
			// define local variables
			unsigned int j;
			evid eventid;
			channel_state state;
			// store struct in local variable
			Admin* ad = admin.at(i);
			// copy event id
			eventid = ad->eventid;
			// get status of channel
			state = ca_state(chanid);
			// set connection status flag
			ad->connectionStatus = state;
			// depending of status from channel
			switch(state) {
			case cs_never_conn: // channel never connected
				// write error message
				wsprintf(ad->sError, "Connection problem for %s", ad->name);
				ad->dError = 6609;
				epicsMutexUnlock(mutex);
				return;
			case cs_prev_conn: // channel was connected, but lost connection
				// check if monitor is running
				if(eventid != 0) {
					// remove "value changed" listener
					iResult = ca_clear_event(eventid);
					// reset eventid
					ad->eventid = 0;
					// check execution of ca_clear_event()
					if(iResult != ECA_NORMAL) {
						wsprintf(ad->sError, "%s occurred while clearing event for %s", ca_message(iResult), ad->name);
						ad->dError = 6605;
					}
				}
				if(ad->eventidEnum != 0) {
					// remove "value changed" listener
					iResult = ca_clear_event(ad->eventidEnum);
					// reset eventid
					ad->eventidEnum = 0;
					// Check execution of ca_clear_event()
					if(iResult != ECA_NORMAL) {
						wsprintf(ad->sError, "%s occurred while clearing event for %s", ca_message(iResult), ad->name);
						ad->dError = 6605;
					}
				}
				// write error message
				wsprintf(ad->sError, "Lost connection for %s", ad->name);
				ad->dError = 6610;
				epicsMutexUnlock(mutex);
				return;
			case cs_conn: // channel is connected
				// get the number of elements from PV
				elems = ca_element_count(chanid);
				ad->elems = elems;
				// identify the field type
				dbrType = ca_field_type(chanid);
				// convert the field type to DBR_TIME_XXXX
				dbrType = dbf_type_to_DBR_TIME(dbrType);
				// store the field type
				ad->dbrType = dbrType;
				// check if the field type is an enum type
				if(dbr_type_is_ENUM(dbrType)) {
					dbrType = DBR_TIME_STRING;
					dbrTypeEnum = DBR_TIME_SHORT;
					ad->enumRecord = 1;
				}
				// clear error in struct
				*(ad->sError) = 0;
				ad->dError = 0;
				// allocate value arrays if channel connected first time
				if(ad->sValueArr == NULL) {
					// allocate string array
					ad->sValueArr = (EpicsStr*)malloc((elems)*sizeof(EpicsStr));
					// check allocation
					if(!ad->sValueArr) {
						wsprintf(ad->sError, "Allocation error for %s", ad->name);
						ad->dError = 6612;
						epicsMutexUnlock(mutex);
						return;
					}
					// set init values
					for(j = 0; j < elems; j++) {
						*(ad->sValueArr[j]) = 0;
					}
				}
				if(ad->dValueArr == NULL) {
					// allocate double array
					ad->dValueArr = (double*)malloc((elems)*sizeof(double));
					// check allocation
					if(!ad->dValueArr) {
						wsprintf(ad->sError, "Allocation error for %s", ad->name);
						ad->dError = 6612;
						epicsMutexUnlock(mutex);
						return;
					}
					// set init values
					for(j = 0; j < elems; j++) {
						ad->dValueArr[j] = std::numeric_limits<double>::quiet_NaN();
					}
				}
				// check if no monitor is running and should start as soon as connection is up again
				if(eventid == 0 && ad->monitor == 1) {
					// add "value changed" listener for PV
					iResult = ca_add_array_event(dbrType, elems, chanid, valueChanged, NULL, 0.0, 0.0, 0.0, &ad->eventid);
					// check execution of ca_add_array_event()
					if(iResult != ECA_NORMAL) {
						wsprintf(ad->sError, "%s occurred while creating event for %s", ca_message(iResult), ad->name);
						ad->dError = 6604;
						epicsMutexUnlock(mutex);
						return;
					}
					// check if its an enum record
					if(ad->enumRecord == 1) {
						// enum - need second monitor for number value
						// add "value changed" listener for PV
						iResult = ca_add_array_event(dbrTypeEnum, elems, chanid, valueChanged, NULL, 0.0, 0.0, 0.0, &ad->eventidEnum);
						// check execution of ca_add_array_event()
						if(iResult != ECA_NORMAL) {
							wsprintf(ad->sError, "%s occurred while creating event for %s", ca_message(iResult), ad->name);
							ad->dError = 6604;
							epicsMutexUnlock(mutex);
							return;
						}
					}
				}
				// check if an array has to put as soon as connection is up again
				if(ad->elemsToPut != 0) {
					// cast dbrType
					switch(ad->dbrType) {
					case DBR_TIME_STRING :
						dbrType = DBR_STRING;
						break;
					case DBR_TIME_SHORT :
						dbrType = DBR_SHORT;
						break;
					case DBR_TIME_FLOAT	:
						dbrType = DBR_FLOAT;
						break;
					case DBR_TIME_ENUM :
						if(ad->enumAsString == 1) {
							dbrType = DBR_STRING;
						} else {
							dbrType = DBR_SHORT;
						}
						break;
					case DBR_TIME_CHAR :
						dbrType = DBR_CHAR;
						break;
					case DBR_TIME_LONG :
						dbrType = DBR_LONG;
						break;
					case DBR_TIME_DOUBLE :
						dbrType = DBR_DOUBLE;
						break;
					}
					// execute ca_array_put
					iResult = ca_array_put(dbrType, ad->elemsToPut, chanid, ad->putArr);
					ca_flush_io();
					if(iResult != ECA_NORMAL) {
						wsprintf(ad->sError, "%s occurred while put %s", ca_message(iResult), ad->name);
						ad->dError = 6617;
						epicsMutexUnlock(mutex);
						return;
					}
					free(ad->putArr);
					ad->putArr = NULL;
					ad->elemsToPut = 0;
				}
				epicsMutexUnlock(mutex);
				return;
			case cs_closed: // connection closed
				// check if monitor is running
				if(eventid != 0) {
					// remove "value changed" listener
					iResult = ca_clear_event(eventid);
					// reset eventid
					ad->eventid = 0;
					// check execution of ca_clear_event()
					if(iResult != ECA_NORMAL) {
						wsprintf(ad->sError, "%s occurred while clearing event for %s", ca_message(iResult), ad->name);
						ad->dError = 6605;
						epicsMutexUnlock(mutex);
						return;
					}
				}
				// write error message
				wsprintf(ad->sError, "Connection closed for %s", ad->name);
				ad->dError = 6611;
				epicsMutexUnlock(mutex);
				return;
			}
		}
	}
	// pv not found
	epicsMutexUnlock(mutex);
	return;
}

//----------------------------------------------------------------------------
// Function:	valueChanged
//
// Description:		Callback function for value events.
//					Updates the structure ('Admin') with current values.
//
// Arg(s) In:	event_handler_args (see CA manual)
//----------------------------------------------------------------------------
void valueChanged(struct event_handler_args args) {
	epicsMutexLock(mutex);
	// check CA Context
	if(!isRunning) {
		epicsMutexUnlock(mutex);
		return;
	}
	// define variables
	unsigned int i; // counter
	chid chanid = args.chid;
	// search pv in vector list
	for(i = 0; i < admin.size(); i++) {
		if(admin.at(i)->chanid == chanid) {
			// found PV
			// store struct in local variable
			Admin* ad = admin.at(i);
			// Get the data pointer from event_handler_args
			const void* data;
			data = args.dbr;
			// Get pointer to the value(s) of pv
			// (i.e. pointer to value field from struct dbr_time_double)
			void* val;
			val = dbr_value_ptr(data, args.type);//ad->dbrType);
			// define local variables
			unsigned int j;
			// copy the informations to administration struct in vector list
			switch(ad->dbrType) {
			case DBR_TIME_STRING :
				// copy value arrays
				for(j = 0; j < ad->elems; j++) {
					strcpy(ad->sValueArr[j], ((dbr_string_t*)val)[j]);
					ad->dValueArr[j] = std::numeric_limits<double>::quiet_NaN();
				}
				break;
			case DBR_TIME_SHORT :
				// copy value arrays
				for(j = 0; j < ad->elems; j++) {
					sprintf(ad->sValueArr[j], "%i", ((dbr_short_t*)val)[j]);
					ad->dValueArr[j] = ((dbr_short_t*)val)[j];
				}
				break;
			case DBR_TIME_FLOAT	:
				// copy value arrays
				for(j = 0; j < ad->elems; j++) {
					sprintf(ad->sValueArr[j], "%g", ((dbr_float_t*)val)[j]);
					ad->dValueArr[j] = ((dbr_float_t*)val)[j];
				}
				break;
			case DBR_TIME_ENUM :
				if(args.type == DBR_TIME_STRING) {
					// copy string value array
					for(j = 0; j < ad->elems; j++) {
						strcpy(ad->sValueArr[j], ((dbr_string_t*)val)[j]);
					}
				}
				if(args.type == DBR_TIME_SHORT) {
					// copy double value array
					for(j = 0; j < ad->elems; j++) {
						ad->dValueArr[j] = ((dbr_short_t*)val)[j];
					}
				}
				break;
			case DBR_TIME_CHAR :
				// copy value arrays
				for(j = 0; j < ad->elems; j++) {
					sprintf(ad->sValueArr[j], "%c", ((dbr_char_t*)val)[j]);
					ad->dValueArr[j] = std::numeric_limits<double>::quiet_NaN();
				}
				break;
			case DBR_TIME_LONG :
				// copy value arrays
				for(j = 0; j < ad->elems; j++) {
					sprintf(ad->sValueArr[j], "%i", ((dbr_long_t*)val)[j]);
					ad->dValueArr[j] = ((dbr_long_t*)val)[j];
				}
				break;
			case DBR_TIME_DOUBLE :
				// copy value arrays
				for(j = 0; j < ad->elems; j++) {
					sprintf(ad->sValueArr[j], "%g", ((dbr_double_t*)val)[j]);
					ad->dValueArr[j] = ((dbr_double_t*)val)[j];
				}
				break;
			}
			// copy status
			strcpy(ad->sStatus, alarmStatusString[((struct dbr_time_short*)data)->status]);
			ad->dStatus = ((struct dbr_time_short*)data)->status;
			// copy severity
			strcpy(ad->sSeverity, alarmSeverityString[((struct dbr_time_short*)data)->severity]); // veraltet
			ad->dSeverity = ((struct dbr_time_short*)data)->severity;
			// copy time
			// convert timestamp to readable string
			epicsTimeToStrftime(ad->sTime, sizeof(ad->sTime), "%Y/%m/%d %H:%M:%S.%03f", &((struct dbr_time_short*)data)->stamp);
			ad->dTime = ((struct dbr_time_short*)data)->stamp.secPastEpoch;
			// done copying
			epicsMutexUnlock(mutex);
			return;
		}
	}
	epicsMutexUnlock(mutex);
	return;
}

//----------------------------------------------------------------------------
// Function:	disconnectVariable
//
// Description:		Disconnect the channel according to passed channel id and
// 					deletes the structure ('Admin').
//
// Arg(s) In:	id - channel id of process variable
//
// Arg(s) Out:	sError - error message
//
// Return:		Error code
//----------------------------------------------------------------------------
int disconnectVariable(	char sError[MAX_ERROR_SIZE],
						long* id) {
	epicsMutexLock(mutex);
	// define local variables
	int iResult = -1;
	int dError = 0;
	unsigned int i; // counter
	chid chanid;
	// cast id to channel id
	chanid = (chid)(*id);
	// search for pv in vector list
	for(i = 0; i < admin.size(); i++) {
		if(admin.at(i)->chanid == chanid) {
			Admin* ad = admin.at(i);
			// check if monitor running
			if(ad->eventid != 0) {
				// remove "value changed" listener
				iResult = ca_clear_event(ad->eventid);
				// Check execution of ca_clear_event()
				if(iResult != ECA_NORMAL) {
					wsprintf(sError, "%s occurred while clearing event for %s", ca_message(iResult), ad->name);
					dError = 6605;
					// no return, because other event and channel should be cleared
				}
			}
			if(ad->eventidEnum != 0) {
				// remove "value changed" listener
				iResult = ca_clear_event(ad->eventidEnum);
				// Check execution of ca_clear_event()
				if(iResult != ECA_NORMAL) {
					wsprintf(sError, "%s occurred while clearing event for %s", ca_message(iResult), ad->name);
					dError = 6605;
					// no return, because channel should be cleared
				}
			}
			iResult = ca_clear_channel(ad->chanid);
			// check execution of ca_clear_channel()
			if(iResult != ECA_NORMAL) {
				wsprintf(sError, "%s occurred while clearing channel for %s", ca_message(iResult), ad->name);
				dError = 6603;
			}
			// TODO: warum fhrt free(...) zu abstrzen in LV 8.5
//			unsigned int j;
//			for(j = 0; j < ad->elems; j++) {
//				free(ad->sValueArr[j]);
//			}
//
//			free(ad->sValueArr);
//			free(ad->dValueArr);
//			free(ad->putArr);
//			free(ad);

			admin.erase(admin.begin()+i, admin.begin()+i+1);
			epicsMutexUnlock(mutex);
			return dError;
		}
	}
	epicsMutexUnlock(mutex);
	return dError;
}

//----------------------------------------------------------------------------
// Function:	diconnectAllVariables
//
// Description:		Disconnect all open channels and
// 					deletes the structures ('Admin')
//
// Return:		Error code
//----------------------------------------------------------------------------
int diconnectAllVariables(char sError[MAX_ERROR_SIZE]) {
	epicsMutexLock(mutex);
	int dError = 0;
	Admin* ad;
	while(admin.size() > 0) {
		ad = admin.front();
		epicsMutexUnlock(mutex);
		if(ad != NULL) {
			dError = disconnectVariable(ad->sError, (long*)&ad->chanid);
		}
		if(dError != 0) {
			strcpy(sError, ad->sError);
		}
		epicsMutexLock(mutex);
	}
	epicsMutexUnlock(mutex);
	return dError;
}

//----------------------------------------------------------------------------
// Function:	addVariableToMonitor
//
// Description:		Starts an monitor event according to passed channel id.
//
// Arg(s) In:	id - the channel id of the process variable
//
// Arg(s) Out:	sError - error message
//
// Return:		Error code
//----------------------------------------------------------------------------
int addVariableToMonitor(	char sError[MAX_ERROR_SIZE],
							long* id ) {
	epicsMutexLock(mutex);
	// check CA Context
	if(!isRunning) {
		wsprintf(sError, "Channel access context not yet ready.");
		epicsMutexUnlock(mutex);
		return 6615;
	}
	// define variables
	int iResult = -1;
	unsigned int i; // counter
	chid chanid;
	// cast id to channel id
	chanid = (chid)(*id);
	// search pv in vector list
	for(i = 0; i < admin.size(); i++) {
		if(admin.at(i)->chanid == chanid) {
			// found PV
			// store struct in local variable
			Admin* ad = admin.at(i);
			// define local variables
			chtype dbrType = ad->dbrType;
			chtype dbrTypeEnum = DBR_TIME_SHORT;
			unsigned short enumRecord = 0;
			unsigned long elems = ad->elems;
			// check if its an enum record
			if(dbrType == DBR_TIME_ENUM) {
				enumRecord = 1;
				dbrType = DBR_TIME_STRING;
				dbrTypeEnum = DBR_TIME_SHORT;
			}
			// flag that monitor should start as soon as the connection is up
			ad->monitor = 1;
			// check if monitor is running
			if(ad->eventid == 0) {
				// check the connection status of the pv
				if(ad->connectionStatus != cs_conn) {
					wsprintf(sError, "No connection to %s. Monitor starts as soon as connection is up again.", ad->name);
					epicsMutexUnlock(mutex);
					return 6616;
				}
				// connection up -> ready for ca_add_array_event
				// add "value changed" listener for PV
				iResult = ca_add_array_event(dbrType, elems, chanid, valueChanged, NULL, 0.0, 0.0, 0.0, &ad->eventid);
				// check execution of ca_add_array_event()
				if(iResult != ECA_NORMAL) {
					wsprintf(sError, "%s occurred while creating event for %s", ca_message(iResult), ad->name);
					epicsMutexUnlock(mutex);
					return 6604;
				}
				// check if its an enum record
				if(enumRecord == 1) {
					// enum - need second monitor for number value
					// add "value changed" listener for PV
					iResult = ca_add_array_event(dbrTypeEnum, elems, chanid, valueChanged, NULL, 0.0, 0.0, 0.0, &ad->eventidEnum);
					// check execution of ca_add_array_event()
					if(iResult != ECA_NORMAL) {
						wsprintf(sError, "%s occurred while creating event for %s", ca_message(iResult), ad->name);
						epicsMutexUnlock(mutex);
						return 6604;
					}
				}
				ca_flush_io();
			}
			epicsMutexUnlock(mutex);
			return 0;
		}
	}
	// pv not found
	epicsMutexUnlock(mutex);
	return 6618;
}

//----------------------------------------------------------------------------
// Function:	removeVariableFromMonitor
//
// Description:		Clears the monitor event according to passed channel id.
//
// Arg(s) In:	id - channel id of process variable
//
// Arg(s) Out:	sError - error message
//
// Return:		Error code
//----------------------------------------------------------------------------
int removeVariableFromMonitor(	char sError[MAX_ERROR_SIZE],
								long* id ) {
	epicsMutexLock(mutex);
	// check CA Context
	if(!isRunning) {
		wsprintf(sError, "Channel access context not yet ready.");
		epicsMutexUnlock(mutex);
		return 6615;
	}
	// define variables
	int iResult = -1;
	unsigned int i; // counter
	chid chanid;
	// cast id to channel id
	chanid = (chid)(*id);
	// search pv in vector list
	for(i = 0; i < admin.size(); i++) {
		if(admin.at(i)->chanid == chanid) {
			// found PV
			// store struct in local variable
			Admin* ad = admin.at(i);
			// flag that monitor should start as soon as the connection is up
			ad->monitor = 0;
			// check if monitor is running
			if(ad->eventid != 0) {
				// remove "value changed" listener for PV
				iResult = ca_clear_event(ad->eventid);
				// Check execution of ca_clear_event()
				if(iResult != ECA_NORMAL) {
					wsprintf(sError, "%s occurred while clearing event for %s", ca_message(iResult), ad->name);
					epicsMutexUnlock(mutex);
					return 6605;
				}
				ad->eventid = 0;
			}
			if(ad->eventidEnum != 0) {
				// remove "value changed" listener for PV
				iResult = ca_clear_event(ad->eventidEnum);
				// Check execution of ca_clear_event()
				if(iResult != ECA_NORMAL) {
					wsprintf(sError, "%s occurred while clearing event for %s", ca_message(iResult), ad->name);
					epicsMutexUnlock(mutex);
					return 6605;
				}
				ad->eventidEnum = 0;
			}
			epicsMutexUnlock(mutex);
			return 0;
		}
	}
	// pv not found
	epicsMutexUnlock(mutex);
	return 6618;
}

//---------------------------------------------------------------------------------------
// Function:	removeVariableFromMonitor
//
// Description:		Copies the data stored in structure ('Admin') to passed parameters.
//
// Arg(s) In:	id - channel id of process variable
//
// Arg(s) Out:	sError - error message
//				dArray - pointer to value array (double)
// 				sStatus - string to store the status info of pv
// 				dStatus - pointer to a short int to store the status info as a number
//				sSeverity - string to store the severity info of pv
// 				dSeverity - pointer to a short int to store the severity info as a number
//				sTime - string to store the time info of pv
// 				dTime - pointer to an int to store the timestamp of pv
//				str - pointer to a string that stores the value array
//
// Return:		Error code
//---------------------------------------------------------------------------------------
int readMonitor(	char sError[MAX_ERROR_SIZE],
					long* id,
					double* dArray,
					char sStatus[MAX_STRING_SIZE],
					short* dStatus,
					char sSeverity[MAX_STRING_SIZE],
					short* dSeverity,
					char sTime[MAX_STRING_SIZE],
					int* dTime,
					char* str ) {

	epicsMutexLock(mutex);
	// check CA Context
	if(!isRunning) {
		wsprintf(sError, "Channel access context not yet ready.");
		epicsMutexUnlock(mutex);
		return 6615;
	}
	// define variables
	unsigned int i; // counter
	chid chanid;
	// cast id to channel id
	chanid = (chid)(*id);
	// search pv in vector list
	for(i = 0; i < admin.size(); i++) {
		if(admin.at(i)->chanid == chanid) {
			// found PV
			// store struct in local variable
			Admin* ad = admin.at(i);
			// define variables
			unsigned int j;
			char number[MAX_STRING_SIZE];
			char* s = str;
			// copy values
			for(j = 0; j < ad->elems; j++) {
				// copy string value array
				strcpy(number, ad->sValueArr[j]);
				// replace the first whitespaces with number
				strncpy(s, number, strlen(number));
				// search for next ';' character
				while(*s != ';') {
					s++;
				}
				s++;
				// copy double value array
				dArray[j] = ad->dValueArr[j];
			}
			// copy status
			strcpy(sStatus, ad->sStatus);
			*dStatus = ad->dStatus;
			// copy severity
			strcpy(sSeverity, ad->sSeverity);
			*dSeverity = ad->dSeverity;
			// copy time
			strcpy(sTime, ad->sTime);
			*dTime = ad->dTime;
			// copy error
			strcpy(sError, ad->sError);

			epicsMutexUnlock(mutex);
			return ad->dError;
		}
	}
	epicsMutexUnlock(mutex);
	return 6618;
}

//--------------------------------------------------------------------------------------------
// Function:	getEgu
//
// Description:		Gets the engineering unit from the pv using an already opened channel
//					and stores it in the structure ('Admin').
//
// Arg(s) In:	id - channel id of process variable
//
// Return:		Error code
//--------------------------------------------------------------------------------------------
int getEgu(char sError[MAX_ERROR_SIZE], long* id) {
	epicsMutexLock(mutex);
	int iResult;
	unsigned int i;
	chid chanid;
	chanid = (chid)(*id);
	// search PV in vector list
	for(i = 0; i < admin.size(); i++) {
		if(admin.at(i)->chanid == chanid) {
			Admin* ad = admin.at(i);
			dbr_gr_short data;
			// concatenate ".EGU" to name of pv
			char eguName[MAX_STRING_SIZE];
			strcpy(eguName, ad->name);
			strcat(eguName, ".EGU");
			// execute ca_get using an already opened channel and a structure that includes the 'units' info
			iResult = ca_get(DBR_GR_SHORT, ad->chanid, (void*)&data);
			if(iResult != ECA_NORMAL) {
				wsprintf(sError, "%s occurred while ca_get for %s", ca_message(iResult), eguName);
				epicsMutexUnlock(mutex);
				return 6607;
			}
			//
			iResult = ca_pend_io(1.0);
			if(iResult != ECA_NORMAL) {
				wsprintf(sError, "%s occurred while ca_get for %s", ca_message(iResult), eguName);
				epicsMutexUnlock(mutex);
				return 6607;
			}
			strcpy(ad->unit, data.units);
			epicsMutexUnlock(mutex);
			return 0;
		}
	}
	epicsMutexUnlock(mutex);
	return 6618;
}

//--------------------------------------------------------------------------------------------
// Function:	getDesc
//
// Description:		Gets the description from the pv and stores it in the structure ('Admin').
//
// Arg(s) In:	id - channel id of process variable
//
// Arg(s) Out:	sError - error message
//
// Return:		Error code
//--------------------------------------------------------------------------------------------
int getDesc(char sError[MAX_ERROR_SIZE], long* id) {
	epicsMutexLock(mutex);
	int iResult;
	unsigned int i;
	chid chanid;
	chanid = (chid)(*id);
	// search PV in vector list
	for(i = 0; i < admin.size(); i++) {
		if(admin.at(i)->chanid == chanid) {
			Admin* ad = admin.at(i);
			chid descChanid;
			dbr_sts_string data;
			// concatenate ".EGU" to name of pv
			char eguName[MAX_STRING_SIZE];
			strcpy(eguName, ad->name);
			strcat(eguName, ".DESC");
			// create channel for 'pvname'.EGU
			iResult = ca_create_channel(eguName, NULL, NULL, CA_PRIORITY_DEFAULT, &descChanid);
			if(iResult != ECA_NORMAL) {
				wsprintf(sError, "%s occurred while creating channel for %s", ca_message(iResult), eguName);
				epicsMutexUnlock(mutex);
				return 6602;
			}
			iResult = ca_pend_io(1.0);
			if(iResult != ECA_NORMAL) {
				wsprintf(sError, "%s occurred while creating channel for %s", ca_message(iResult), eguName);
				epicsMutexUnlock(mutex);
				return 6602;
			}
			iResult = ca_get(DBR_STS_STRING, descChanid, (void*)&data);
			if(iResult != ECA_NORMAL) {
				wsprintf(sError, "%s occurred while ca_get for %s", ca_message(iResult), eguName);
				epicsMutexUnlock(mutex);
				return 6607;
			}
			iResult = ca_pend_io(1.0);
			if(iResult != ECA_NORMAL) {
				wsprintf(sError, "%s occurred while ca_get for %s", ca_message(iResult), eguName);
				epicsMutexUnlock(mutex);
				return 6607;
			}
			strcpy(ad->desc, data.value);
			epicsMutexUnlock(mutex);
			return 0;
		}
	}
	epicsMutexUnlock(mutex);
	return 6618;
}

//--------------------------------------------------------------------------------------------
// Function:	getElemsAndType
//
// Description:		Copies the number of elements and the dbr type of pv to passed parameters.
//
// Arg(s) In:	id - channel id of process variable
//
// Arg(s) Out:	sError - error message
//				elems - pointer to an unsigned long int to store the number of elements of pv
//				dbrType - string to store the dbr type of pv
//
// Return:		Error code
//--------------------------------------------------------------------------------------------
int getElemsAndType(	char sError[MAX_ERROR_SIZE],
						long* id,
						unsigned long* elems,
						char dbrType[MAX_ERROR_SIZE]) {
	epicsMutexLock(mutex);
	// check CA Context
	if(!isRunning) {
		wsprintf(sError, "Channel access context not yet ready.");
		epicsMutexUnlock(mutex);
		return 6615;
	}
	// define variables
	unsigned int i = 0; // counter
	chid chanid;
	// cast id to channel id
	chanid = (chid)(*id);
	// search pv in vector list
	for(i = 0; i < admin.size(); i++) {
		if(admin.at(i)->chanid == chanid) {
			// found PV
			// store struct in local variable
			Admin* ad = admin.at(i);
			// copy number of elements
			*elems = ad->elems;
			// copy dbrType
			switch(ad->dbrType) {
			case DBR_TIME_STRING :
				strcpy(dbrType, "DBR_STRING");
				break;
			case DBR_TIME_SHORT :
				strcpy(dbrType, "DBR_SHORT");
				break;
			case DBR_TIME_FLOAT	:
				strcpy(dbrType, "DBR_FLOAT");
				break;
			case DBR_TIME_ENUM :
				strcpy(dbrType, "DBR_ENUM");
				break;
			case DBR_TIME_CHAR :
				strcpy(dbrType, "DBR_CHAR");
				break;
			case DBR_TIME_LONG :
				strcpy(dbrType, "DBR_LONG");
				break;
			case DBR_TIME_DOUBLE :
				strcpy(dbrType, "DBR_DOUBLE");
				break;
			case TYPENOTCONN :
				printf("channel disconnected\n");
				break;
			default :
				printf("no valid dbr_type\n");
				break;
			}
			// clear error
			*sError = 0;

			epicsMutexUnlock(mutex);
			return 0;
		}
	}
	// pv not found (should never happen)
	*elems = 0;
	// write error
	wsprintf(sError, "getNumberOfElements: Can not find pv");
	epicsMutexUnlock(mutex);
	return 6618;
}

//--------------------------------------------------------------------------------------------
// Function:	buildOverviewArray
//
// Description:		Copies name, connection info and the dbr type of pv to passed parameters
//					to build an overview in LabVIEW.
//
// Arg(s) In:	id - channel id of process variable
// 				count - a counter that starts by zero and increments by one
//
// Arg(s) Out:	sError - error message
// 				pvname - string to store the name of pv
// 				connected - pointer to a short int to indicate whether pv is connected or not (bool)
//				dbrType - string to store the dbr type of pv
//				end - pointer to an unsigned short int to indicate that all pvs got visited
//
// Return:		Error code
//--------------------------------------------------------------------------------------------
int buildOverviewArray(	char sError[MAX_ERROR_SIZE],
						char pvname[MAX_STRING_SIZE],
						long* id,
						char unit[MAX_UNITS_SIZE],
						char desc[MAX_STRING_SIZE],
						short* connected,
						char dbrType[MAX_STRING_SIZE],
						unsigned short* count,
						unsigned short* end ) {

	epicsMutexLock(mutex);
	// check CA Context
	if(!isRunning) {
		wsprintf(sError, "Channel access context not yet ready.");
		epicsMutexUnlock(mutex);
		return 6615;
	}
	if(admin.size() > *count) {
		Admin* ad = admin.at(*count);
		if(ad != NULL) {
			strcpy(pvname, ad->name);
			*id = (long)ad->chanid;
			if(ad->connectionStatus == cs_conn) {
				*connected = 1;
			} else {
				*connected = 0;
			}
			strcpy(unit, ad->unit);
			strcpy(desc, ad->desc);
			switch(ad->dbrType) {
			case DBR_TIME_STRING :
				strcpy(dbrType, "DBR_STRING");
				break;
			case DBR_TIME_SHORT :
				strcpy(dbrType, "DBR_SHORT");
				break;
			case DBR_TIME_FLOAT	:
				strcpy(dbrType, "DBR_FLOAT");
				break;
			case DBR_TIME_CHAR :
				strcpy(dbrType, "DBR_CHAR");
				break;
			case DBR_TIME_LONG :
				strcpy(dbrType, "DBR_LONG");
				break;
			case DBR_TIME_DOUBLE :
				strcpy(dbrType, "DBR_DOUBLE");
				break;
			case TYPENOTCONN :
				printf("channel disconnected\n");
				break;
			default :
				printf("no valid dbr_type\n");
				break;
			}
		}
		epicsMutexUnlock(mutex);
		return 0;
	}
	*end = 1;
	epicsMutexUnlock(mutex);
	return 0;
}

//--------------------------------------------------------------------------------------------
// Function:	caGet
//
// Description:		Gets current value(s) and status infos from pv
// 					and copies them to passed parameters.
//
// Arg(s) In:	id - channel id of process variable
//
// Arg(s) Out:	sError - error message
//				dArray - pointer to value array (double)
// 				sStatus - string to store the status info of pv
// 				dStatus - pointer to a short int to store the status info as a number
//				sSeverity - string to store the severity info of pv
// 				dSeverity - pointer to a short int to store the severity info as a number
//				sTime - string to store the time info of pv
// 				dTime - pointer to an int to store the timestamp of pv
//				str - pointer to a string that stores the value array
//
// Return:		Error code
//--------------------------------------------------------------------------------------------
int caGet(	char sError[MAX_ERROR_SIZE],
			long* id,
			double* dArray,
			char sStatus[MAX_STRING_SIZE],
			short* dStatus,
			char sSeverity[MAX_STRING_SIZE],
			short* dSeverity,
			char sTime[MAX_STRING_SIZE],
			int* dTime,
			char* str) {

	epicsMutexLock(mutex);
	// check CA Context
	if(!isRunning) {
		wsprintf(sError, "Channel access context not yet ready.");
		epicsMutexUnlock(mutex);
		return 6615;
	}
	// define variables
	int iResult = -1;
	int status = -1;
	unsigned int i; // counter
	chid chanid;
	// cast id to channel id
	chanid = (chid)(*id);
	// search pv in vector list
	for(i = 0; i < admin.size(); i++) {
		if(admin.at(i)->chanid == chanid) {
			// found PV
			// store struct in local variable
			Admin* ad = admin.at(i);
			// define local variables
			unsigned int j;
			chtype dbrType = ad->dbrType;
			chtype dbrTypeEnum = DBR_TIME_SHORT;
			unsigned short enumRecord = 0; // 0 - no, 1 - enum record
			unsigned long elems = ad->elems;
			char number[MAX_STRING_SIZE];
			char* s = str;
			epicsTimeStamp currentTime;
			double timeDiff = 0;
			// check if its an enum record
			if(dbrType == DBR_TIME_ENUM) {
				enumRecord = 1;
				dbrType = DBR_TIME_STRING;
				dbrTypeEnum = DBR_TIME_SHORT;
			}
			// check the connection status of the pv
			if(ad->connectionStatus != cs_conn) {
				// set to last known values
				if(ad->sValueArr != NULL && ad->dValueArr != NULL) {
					// copy values
					for(j = 0; j < elems; j++) {
						// copy string value array
						strcpy(number, ad->sValueArr[j]);
						// replace the first whitespaces with number
						strncpy(s, number, strlen(number));
						// search for next ';' character
						while(*s != ';') {
							s++;
						}
						s++;
						// copy double value array
						dArray[j] = ad->dValueArr[j];
					}
				}
				// copy status
				strcpy(sStatus, ad->sStatus);
				*dStatus = ad->dStatus;
				// copy severity
				strcpy(sSeverity, ad->sSeverity);
				*dSeverity = ad->dSeverity;
				// copy time
				strcpy(sTime, ad->sTime);
				*dTime = ad->dTime;
				// write error
				wsprintf(sError, "Connection problem for %s", ad->name);

				epicsMutexUnlock(mutex);
				return 6609;
			}
			// calculate time difference between last and current caget
			if(ad->timestamp.nsec != 0 && ad->timestamp.secPastEpoch != 0) {
				epicsTimeGetCurrent(&currentTime);
				timeDiff = epicsTimeDiffInSeconds(&currentTime, &ad->timestamp);
			}
			// check if a monitor is running
			if(ad->eventid != 0 ||
			  (timeDiff < REFRESH_RATE && ad->timestamp.nsec != 0 && ad->timestamp.secPastEpoch != 0)) {
				// monitor is running or
				// time difference between two cagets is too small
				// -> copy current information from stored data
				// copy string value array
				// copy values
				for(j = 0; j < elems; j++) {
					// copy string value array
					strcpy(number, ad->sValueArr[j]);
					// replace the first whitespaces with number
					strncpy(s, number, strlen(number));
					// search for next ';' character
					while(*s != ';') {
						s++;
					}
					s++;
					// copy double value array
					dArray[j] = ad->dValueArr[j];
				}
				// copy status
				strcpy(sStatus, ad->sStatus);
				*dStatus = ad->dStatus;
				// copy severity
				strcpy(sSeverity, ad->sSeverity);
				*dSeverity = ad->dSeverity;
				// copy time
				strcpy(sTime, ad->sTime);
				*dTime = ad->dTime;
				// copy error
				strcpy(sError, ad->sError);

				epicsMutexUnlock(mutex);
				return ad->dError;
			}
			// connection up -> no current data available -> ready for ca_array_get
			// get number of bytes for time field structure (data request buffer)
			unsigned long nBytes = dbr_size_n(dbrType, elems);
			unsigned long nBytesEnum = 0;
			if(enumRecord == 1) {
				nBytesEnum = dbr_size_n(dbrTypeEnum, elems);
			}
			// define pointer to time field structure
			void* data = malloc(nBytes);
			if(!data) { // allocation error
				wsprintf(sError, "Allocation error");
				epicsMutexUnlock(mutex);
				return 6612;
			}
			void* dataEnum = NULL;
			if(enumRecord == 1) {
				dataEnum = malloc(nBytesEnum);
				if(!dataEnum) { // allocation error
					wsprintf(sError, "Allocation error");
					epicsMutexUnlock(mutex);
					return 6612;
				}
			}

			// execute ca_array_get
			iResult = ca_array_get(dbrType, elems, chanid, data);
			// wait 1 sec for execution of ca_array_get
			status = ca_pend_io(1.0);
			// check execution of ca_pend_io()
			if(status != ECA_NORMAL) {
				wsprintf(sError, "%s occurred while creating event for %s", ca_message(iResult), ad->name);
				epicsMutexUnlock(mutex);
				return 6604;
			}
			// check the execution of ca_array_get
			if(iResult != ECA_NORMAL) {
				wsprintf(sError, "%s occurred while ca_get for %s", ca_message(iResult), ad->name);
				epicsMutexUnlock(mutex);
				return 6607;
			}

			// check if its an enum record
			if(enumRecord == 1) {
				// enum - need second caget for number value
				// execute ca_array_get
				iResult = ca_array_get(dbrTypeEnum, elems, chanid, dataEnum);
				// wait 1 sec for execution of ca_array_get
				status = ca_pend_io(1.0);
				// check execution of ca_pend_io()
				if(status != ECA_NORMAL) {
					wsprintf(sError, "%s occurred while creating event for %s", ca_message(iResult), ad->name);
					epicsMutexUnlock(mutex);
					return 6604;
				}
				// check the execution of ca_array_get
				if(iResult != ECA_NORMAL) {
					wsprintf(sError, "%s occurred while ca_get for %s", ca_message(iResult), ad->name);
					epicsMutexUnlock(mutex);
					return 6607;
				}

			}
			// set time
			// TODO: ad->timstamp.secPastEpoch = currentTime.secPastEpoch;
			// TODO: ad->timestamp.nsec = currentTime.nsec;
			epicsTimeGetCurrent(&ad->timestamp);
			// define pointer to the begin of the value array
			void* val;
			void* valEnum = NULL;
			// get pointer to the begin of the value array stored in time field structure (data)
			val = dbr_value_ptr(data, dbrType);
			if(enumRecord == 1) {
				valEnum = dbr_value_ptr(dataEnum, dbrTypeEnum);
			}
			// copy the informations to the given parameter
			switch(dbrType) {
			case DBR_TIME_STRING :
				// copy values
				for(j = 0; j < elems; j++) {
					// copy string value array
					sprintf(number, "%s", ((dbr_string_t*)val)[j]);
					// replace the first whitespaces with number
					strncpy(s, number, strlen(number));
					// search for next ';' character
					while(*s != ';') {
						s++;
					}
					s++;
					// copy to struct ('Admin')
					strcpy(ad->sValueArr[j], number);
					// copy double value array
					// check if its an enum record
					if(enumRecord == 1) {
						// copy enum number values
						dArray[j] = ((dbr_short_t*)valEnum)[j];
						ad->dValueArr[j] = ((dbr_short_t*)valEnum)[j];
					} else {
						// no enum (val = NAN)
						// copy double value array
						dArray[j] = std::numeric_limits<double>::quiet_NaN();
						ad->dValueArr[j] = std::numeric_limits<double>::quiet_NaN();
					}
				}
				break;
			case DBR_TIME_SHORT :
				// copy values
				for(j = 0; j < elems; j++) {
					// copy string value array
					sprintf(number, "%i", ((dbr_short_t*)val)[j]);
					// replace the first whitespaces with number
					strncpy(s, number, strlen(number));
					// search for next ';' character
					while(*s != ';') {
						s++;
					}
					s++;
					// copy double value array
					dArray[j] = ((dbr_short_t*)val)[j];
					// copy to struct ('Admin')
					strcpy(ad->sValueArr[j], number);
					ad->dValueArr[j] = ((dbr_short_t*)val)[j];
				}
				break;
			case DBR_TIME_FLOAT	:
				// copy values
				for(j = 0; j < elems; j++) {
					// copy string value array
					sprintf(number, "%g", ((dbr_float_t*)val)[j]);
					// replace the first whitespaces with number
					strncpy(s, number, strlen(number));
					// search for next ';' character
					while(*s != ';') {
						s++;
					}
					s++;
					// copy double value array
					dArray[j] = ((dbr_float_t*)val)[j];
					// copy to struct ('Admin')
					strcpy(ad->sValueArr[j], number);
					ad->dValueArr[j] = ((dbr_float_t*)val)[j];
				}
				break;
			case DBR_TIME_CHAR :
				// copy values
				for(j = 0; j < elems; j++) {
					// copy string value array
					sprintf(number, "%c", ((dbr_char_t*)val)[j]);
					// replace the first whitespaces with number
					strncpy(s, number, strlen(number));
					// search for next ';' character
					while(*s != ';') {
						s++;
					}
					s++;
					// copy double value array
					dArray[j] = std::numeric_limits<double>::quiet_NaN();
					// copy to struct ('Admin')
					strcpy(ad->sValueArr[j], number);
					ad->dValueArr[j] = std::numeric_limits<double>::quiet_NaN();
				}
				break;
			case DBR_TIME_LONG :
				// copy values
				for(j = 0; j < elems; j++) {
					// copy string value array
					sprintf(number, "%i", ((dbr_long_t*)val)[j]);
					// replace the first whitespaces with number
					strncpy(s, number, strlen(number));
					// search for next ';' character
					while(*s != ';') {
						s++;
					}
					s++;
					// copy double value array
					dArray[j] = ((dbr_long_t*)val)[j];
					// copy to struct ('Admin')
					strcpy(ad->sValueArr[j], number);
					ad->dValueArr[j] = ((dbr_long_t*)val)[j];
				}
				break;
			case DBR_TIME_DOUBLE :
				// copy values
				for(j = 0; j < elems; j++) {
					// copy string value array
					sprintf(number, "%g", ((dbr_double_t*)val)[j]);
					// replace the first whitespaces with number
					strncpy(s, number, strlen(number));
					// search for next ';' character
					while(*s != ';') {
						s++;
					}
					s++;
					// copy double value array
					dArray[j] = ((dbr_double_t*)val)[j];
					// copy to struct ('Admin')
					strcpy(ad->sValueArr[j], number);
					ad->dValueArr[j] = ((dbr_double_t*)val)[j];
				}
				break;
			case TYPENOTCONN : // the channel's native type when disconnected
				wsprintf(sError, "Connection problem for %s ", ad->name);
				epicsMutexUnlock(mutex);
				return 6609;
			default :
				wsprintf(sError, "No valid DBR_TYPE for %s", ad->name);
				epicsMutexUnlock(mutex);
				return 6621;
			}
			// cast data
			dbr_time_short* dataB = (struct dbr_time_short*)data;
			// copy status
			strcpy(sStatus, alarmStatusString[dataB->status]);
			strcpy(ad->sStatus, alarmStatusString[dataB->status]);
			*dStatus = dataB->status;
			ad->dStatus = dataB->status;
			// copy severity
			strcpy(sSeverity, alarmSeverityString[dataB->severity]); // veraltet
			strcpy(ad->sSeverity, alarmSeverityString[dataB->severity]); // veraltet
			*dSeverity = dataB->severity;
			ad->dSeverity = dataB->severity;
			// copy time
			epicsTimeToStrftime(number, sizeof(number), "%Y/%m/%d %H:%M:%S.%03f", &dataB->stamp);
			strcpy(sTime, number);
			strcpy(ad->sTime, sTime);
			*dTime = ((struct dbr_time_double*)data)->stamp.secPastEpoch;
			ad->dTime = ((struct dbr_time_double*)data)->stamp.secPastEpoch;
			// done copying
			// free allocated memory
			free(data);
			free(dataEnum);

			epicsMutexUnlock(mutex);
			return 0;
		}
	}
	// pv not found (should never happen)
	epicsMutexUnlock(mutex);
	return 6618;
}

//--------------------------------------------------------------------------------------------
// Function:	caPut
//
// Description:		Writes passed value(s) to a process variable.
//
// Arg(s) In:	id - channel id of process variable
// 				elems - number of elements to write
// 				enumAsString - indicates whether an enum variable should write as a text (string)
//							   or as a number (bool)
// 				arr - a pointer to an array that stores the values that should be written
// 				str - a string that stores the values that should be written
//
// Return:		Error code
//--------------------------------------------------------------------------------------------
int caPut(	char sError[MAX_ERROR_SIZE],
			long* id,
			unsigned long* elems,
			unsigned short* enumAsString,
			void* arr,
			char* str) {

	epicsMutexLock(mutex);
	// check CA Context
	if(!isRunning) {
		wsprintf(sError, "Channel access context not yet ready.");
		epicsMutexUnlock(mutex);
		return 6615;
	}
	// define variables
	int iResult = -1;
	unsigned int i; // counter
	chid chanid;
	double* dblArr = (double*)arr;
	long* longArr;
	short* shortArr;
	float* floatArr;
	EpicsStr* strArr;
	char* chrArr;
	// cast id to channel id
	chanid = (chid)(*id);
	// search pv in vector list
	for(i = 0; i < admin.size(); i++) {
		if(admin.at(i)->chanid == chanid) {
			// found PV
			// store struct in local variable
			Admin* ad = admin.at(i);
			// define local variables
			unsigned int j, stepCnt; // counter
			char* s = str;
			char* tmpStr = s;
			chtype dbrType = DBR_DOUBLE;
			// prepare the array to be written
			switch(ad->dbrType) {
			case DBR_TIME_STRING :
				// copy string to string array
				// allocate string array
				strArr = (EpicsStr*)malloc(*elems * sizeof(EpicsStr));
				memset(strArr, 0, sizeof(*elems * sizeof(EpicsStr)));
				// check allocation
				if(!strArr) {
					wsprintf(sError, "Allocation error");
					epicsMutexUnlock(mutex);
					return 6612;
				}
				if(*str != 0) { // use str to convert to sting array
					// copy values
					for(j = 0; j < *elems; j++) {
						stepCnt = 0;
						tmpStr = s;
						// search for next ';' character
						while(*tmpStr != ';') {
							tmpStr++;
							stepCnt++;
						}
						// copy to string value array
						strncpy(strArr[j], s, stepCnt);
						strArr[j][stepCnt] = '\0';
						tmpStr++;
						s = tmpStr;
					}
				} else { // use double array to convert to string array
					for(j = 0; j < *elems; j++) {
						sprintf(strArr[j], "%g", dblArr[j]);
					}
				}
				arr = strArr;
				ad->putArr = strArr;
				dbrType = DBR_STRING;
				break;
			case DBR_TIME_SHORT :
				shortArr = (short*)malloc((*elems)*sizeof(short));
				// check allocation
				if(!shortArr) {
					wsprintf(sError, "Allocation error");
					epicsMutexUnlock(mutex);
					return 6612;
				}
				for(j = 0; j < *elems; j++) {
					shortArr[j] = (short)dblArr[j];
				}
				arr = shortArr;
				ad->putArr = shortArr;
				dbrType = DBR_SHORT;
				break;
			case DBR_TIME_FLOAT	:
				floatArr = (float*)malloc((*elems)*sizeof(float));
				// check allocation
				if(!floatArr) {
					wsprintf(sError, "Allocation error");
					epicsMutexUnlock(mutex);
					return 6612;
				}
				for(j = 0; j < *elems; j++) {
					floatArr[j] = (float)dblArr[j];
				}
				arr = floatArr;
				ad->putArr = floatArr;
				dbrType = DBR_FLOAT;
				break;
			case DBR_TIME_ENUM :
				// decide if enum put as string or as number
				if(*enumAsString == 1) {
					// copy string to string array
					// allocate string array
					strArr = (EpicsStr*)malloc(*elems * sizeof(EpicsStr));
					memset(strArr, 0, sizeof(*elems * sizeof(EpicsStr)));
					// check allocation
					if(!strArr) {
						wsprintf(sError, "Allocation error");
						epicsMutexUnlock(mutex);
						return 6612;
					}
					// copy values
					for(j = 0; j < *elems; j++) {
						stepCnt = 0;
						tmpStr = s;
						// search for next ';' character
						while(*tmpStr != ';') {
							tmpStr++;
							stepCnt++;
						}
						// copy to string value array
						strncpy(strArr[j], s, stepCnt);
						strArr[j][stepCnt] = '\0';
						tmpStr++;
						s = tmpStr;
					}
					arr = strArr;
					ad->putArr = strArr;
					dbrType = DBR_STRING;
				} else {
					shortArr = (short*)malloc((*elems)*sizeof(short));
					// check allocation
					if(!shortArr) {
						wsprintf(sError, "Allocation error");
						epicsMutexUnlock(mutex);
						return 6612;
					}
					for(j = 0; j < *elems; j++) {
						shortArr[j] = (short)dblArr[j];
					}
					arr = shortArr;
					ad->putArr = shortArr;
					dbrType = DBR_SHORT;
				}
				break;
			case DBR_TIME_CHAR :
				// copy string to string array
				// allocate string array
				chrArr = (char*)malloc((*elems)*sizeof(char));
				// check allocation
				if(!chrArr) {
					wsprintf(sError, "Allocation error");
					epicsMutexUnlock(mutex);
					return 6612;
				}
				for(j = 0; j < *elems; j++) {
					chrArr[j] = *s;
					while(*s != ';' && *s != '\0') {
						s++;
					}
					s++;
				}
				arr = chrArr;
				ad->putArr = chrArr;
				dbrType = DBR_CHAR;
				break;
			case DBR_TIME_LONG :
				longArr = (long*)malloc((*elems)*sizeof(long));
				// check allocation
				if(!longArr) {
					wsprintf(sError, "Allocation error");
					epicsMutexUnlock(mutex);
					return 6612;
				}
				for(j = 0; j < *elems; j++) {
					longArr[j] = (long)dblArr[j];
				}
				arr = longArr;
				ad->putArr = longArr;
				dbrType = DBR_LONG;
				break;
			case DBR_TIME_DOUBLE :
				arr = dblArr;
				ad->putArr = dblArr;
				dbrType = DBR_DOUBLE;
				break;
			}
			// check the connection status of the pv
			if(ad->connectionStatus != cs_conn) {
				// copy array and put it as soon as connection is up again
				// (put in connectionHandler)
				ad->enumAsString = *enumAsString;
				ad->elemsToPut = *elems;
				// write error
				wsprintf(sError, "Connection problem for %s. Values will be written as soon as connection is up again.", ad->name);
				epicsMutexUnlock(mutex);
				return 6620;
			}
			// connection up -> ready for ca_array_put
			// execute ca_array_put
			iResult = ca_array_put(dbrType, *elems, chanid, arr);
			ca_flush_io();
			if(iResult != ECA_NORMAL) {
				wsprintf(sError, "%s occurred while put %s", ca_message(iResult), ad->name);
				epicsMutexUnlock(mutex);
				return 6617;
			}
			epicsMutexUnlock(mutex);
			return 0;
		}
	}

	// pv not found (should never happen)
	epicsMutexUnlock(mutex);
	return 6618;
}

// initialization and cleanup of library
extern "C" {
	BOOL WINAPI DllMain (HANDLE hDll, DWORD dwReason, LPVOID lpReserved) {
		int iResult = -1;
		switch (dwReason) {
			// create CA context if library has been attached
			case DLL_PROCESS_ATTACH:
				dllLoad++;
				isRunning = false;
				mutex = epicsMutexMustCreate();
				iResult = ca_context_create(ca_enable_preemptive_callback);
				if (iResult != ECA_NORMAL) {
					// CA error occurred
					break;
				}
				// channel access context created
				isRunning = true;
				// DLL is loaded
				break;
			// remove CA context if library has been detached
			case DLL_PROCESS_DETACH:
				dllLoad--;
				if(dllLoad == 0) { // DLL is not in use
					isRunning = false;
					char sError[MAX_ERROR_SIZE];
					diconnectAllVariables(sError);
					epicsMutexLock(mutex);
					admin.clear();
					epicsMutexUnlock(mutex);
					ca_context_destroy();
					// channel access context destroyed
					// DLL is freed
				}
				break;
			case DLL_THREAD_ATTACH:
				//cout << "thread is created during the DLL's lifetime" << endl << flush;
				break;

			case DLL_THREAD_DETACH:
				//cout << "thread ends normally" << endl << flush;
				break;
		}
		return TRUE;
	}
}
